import numpy as np


def stack_complex(unstacked: np.ndarray) -> np.ndarray:
    """ Stack a complex array

    The input is an nD array (n > 1) containing complex data.

    Returns a stacked array where the last dimension is double the input last dimension:
        * first half is all the real data
        * second half is all the imaginary data

    """
    assert unstacked.ndim > 1

    shape = list(unstacked.shape)
    shape[-1] = shape[-1] * 2
    stacked = np.empty(shape, dtype=np.complex64)
    half = unstacked.shape[-1]
    stacked[..., :half] = np.real(unstacked)
    stacked[..., half:] = np.imag(unstacked)

    return stacked


def unstack_complex(stacked: np.ndarray) -> np.ndarray:
    """ Unstack a stacked complex array

    The input is an nD array (n > 1) where the last dimension contains stacked complex data:
        * first half is all the real data
        * second half is all the imaginary data

    Returns an unstacked complex array.

    """
    assert stacked.ndim > 1
    assert stacked.shape[-1] % 2 == 0

    half = stacked.shape[-1] // 2
    unstacked = 1j * stacked[..., half:]
    unstacked += stacked[..., :half]

    return unstacked


def stacked_complex_real(stacked: np.ndarray) -> np.ndarray:
    """ Get the real elements from a stacked complex array

    """
    assert stacked.ndim > 1
    assert stacked.shape[-1] % 2 == 0

    half = stacked.shape[-1] // 2
    return stacked[..., :half]


def stacked_complex_imag(stacked: np.ndarray) -> np.ndarray:
    """ Get the imaginary elements from a stacked complex array

    """
    assert stacked.ndim > 1
    assert stacked.shape[-1] % 2 == 0

    half = stacked.shape[-1] // 2
    return stacked[..., half:]



import numpy as np
from pyaaware import ForwardTransform
from pyaaware import InverseTransform

from sonusai.mixture.types import AudioF
from sonusai.mixture.types import AudioT


def get_next_noise(audio: AudioT, offset: int, length: int) -> AudioT:
    return np.take(audio, range(offset, offset + length), mode='wrap')


def read_audio(name: str) -> AudioT:
    import sox

    from sonusai import SonusAIError
    from sonusai import logger
    from sonusai.mixture import tokenized_expandvars
    from sonusai.mixture import BIT_DEPTH
    from sonusai.mixture import CHANNEL_COUNT
    from sonusai.mixture import SAMPLE_RATE

    expanded_name, _ = tokenized_expandvars(name)

    try:
        # Read in and convert to desired format
        inp = sox.Transformer()
        inp.set_output_format(rate=SAMPLE_RATE, bits=BIT_DEPTH, channels=CHANNEL_COUNT)
        return inp.build_array(input_filepath=expanded_name,
                               sample_rate_in=int(sox.file_info.sample_rate(expanded_name)))

    except Exception as e:
        if name != expanded_name:
            logger.error(f'Error reading {name} (expanded: {expanded_name}): {e}')
        else:
            raise SonusAIError(f'Error reading {name}: {e}')


def calculate_transform_from_audio(audio: AudioT, transform: ForwardTransform) -> AudioF:
    """ Apply forward transform to input audio data to generate transform data

    audio
        Input data [samples]
    transform
        ForwardTransform object
    """
    return transform.execute_all(audio).transpose()


def calculate_audio_from_transform(data: AudioF, transform: InverseTransform, trim: bool = True) -> AudioT:
    """ Apply inverse transform to input transform data to generate audio data

    audio
        Input data [frames, bins]
    transform
        InverseTransform object
    trim
        Removes starting samples so output waveform will be time-aligned with input waveform to the transform
    """
    audio = transform.execute_all(data.transpose())
    if trim:
        audio = audio[transform.N - transform.R:]

    return audio



"""sonusai calc_metric_spenh_targetf

usage: calc_metric_spenh_targetf [-hvptx] [-i MIXID] (-d PLOC) INPUT

options:
    -h, --help
    -v, --verbose               Be verbose.
    -i MIXID, --mixid MIXID     Mixture ID(s) to generate. [default: *].
    -d PLOC, --ploc PLOC        Location of SonusAI predict data.
    -p, --plot-enable           Generate PDF file of plots for each mixture.
    -t, --truth-est-mode        Calculate extraction using truth and include metrics.
    -x, --disable-wav           Disable WAV file generation.

Calculate speech enhancement metrics for prediction data in PLOC run with feature+truth data in INPUT.

Inputs:
    PLOC    SonusAI prediction data directory.
    INPUT   SonusAI mixture database directory containing feature+truth data.

Outputs the following to PLOC:
    <id>-target.wav
    <id>-target-est.wav
    <id>-spenh-metrics.txt
    <id>-noise.wav
    <id>-noise-est.wav
    <id>-mixture.wav
    <id>-spenh-plots.pdf
    spenh-metrics-estats-list.csv
    spenh-metrics-estats-tru-list.csv
    spenh-metrics-list.csv
    spenh-metrics-summary.csv
    spenh-metrics-summary.txt
    spenh-metrics-tru-list.csv

"""
from typing import Union

import matplotlib.pyplot as plt
import numpy as np

from sonusai import SonusAIError
from sonusai import logger
from sonusai.mixture import AudioF
from sonusai.mixture import AudioT
from sonusai.mixture import Feature
from sonusai.mixture import Predict


def mean_square_error(ytrue: np.ndarray, ypred: np.ndarray, squared=False) -> (np.ndarray, np.ndarray, np.ndarray):
    """ Calculate mean square error or root mean square error between ytrue, ypred

    Typical inputs expected to be size frames x classes/bins
    returns:
      err    mean over both dimensions
      err_c  mean-over-dim0 i.e. a value per class/bin
      err_f  mean-over-dim1 i.e. a value per frame
    """
    sqerr = np.square(ytrue - ypred)
    err_c = np.mean(sqerr, axis=0)  # mean over frames for value per class
    err_f = np.mean(sqerr, axis=1)  # mean over classes for value per frame
    err = np.mean(sqerr)  # mean over all
    if not squared:
        err_c = np.sqrt(err_c)
        err_f = np.sqrt(err_f)
        err = np.sqrt(err)

    return err, err_c, err_f


def mean_abs_percentage_error(ytrue: np.ndarray, ypred: np.ndarray) -> (np.ndarray, np.ndarray, np.ndarray):
    """ Calculate mean abs percentage error between ytrue, ypred

    Typical inputs expected to be size frames x classes/bins
    returns:
      err    mean over both dimensions
      err_c  mean-over-dim0 i.e. a value per class/bin
      err_f  mean-over-dim1 i.e. a value per frame
    """
    abserr = np.abs(ytrue - ypred) / (ytrue + np.finfo(np.float32).eps)
    err_c = np.mean(abserr, axis=0)  # mean over frames for value per class
    err_f = np.mean(abserr, axis=1)  # mean over classes for value per frame
    err = np.mean(abserr)  # mean over all

    return err, err_c, err_f


def log_error(ytrue: np.ndarray, ypred: np.ndarray) -> (np.ndarray, np.ndarray, np.ndarray):
    """ Calculate log error between ytrue, ypred which are complex or real float numpy arrays

    Typical inputs expected to be size frames x classes/bins
    returns:
      logerr    mean over both dimensions, scalar value
      logerr_c  mean-over-dim0 i.e. a value per class/bin
      logerr_f  mean-over-dim1 i.e. a value per frame
    """
    ytsq = np.real(ytrue * np.conjugate(ytrue))
    ypsq = np.real(ypred * np.conjugate(ypred))
    lgerr = abs(10 * np.log10((ytsq + np.finfo(np.float32).eps) / (ypsq + np.finfo(np.float32).eps)))
    # lgerr = abs(10*np.log10(ytsq / (ypsq + np.finfo(np.float32).eps)+np.finfo(np.float32).eps))
    logerr_c = np.mean(lgerr, axis=0)  # mean over frames for value per class
    logerr_f = np.mean(lgerr, axis=1)  # mean over classes for value per frame
    logerr = np.mean(lgerr)

    return logerr, logerr_c, logerr_f


def calculate_pesq(speech_ref: np.ndarray, speech_est: np.ndarray, error_value=0.0):
    """ Computes the PESQ score of speech estimate audio vs. the clean speech estimate audio

    Upon error, assigns a value of 0, or user specified value in error_value
    Argument/s:
      speech_ref - speech reference audio, np.ndarray vector of samples.
      speech_est - speech estimated audio, np.ndarray vector of samples.
    Returns:
      score - float value between -0.5 to 4.5.
   """
    from pesq import pesq

    try:
        score = pesq(16000, speech_ref, speech_est, mode='wb')
    except Exception as e:
        logger.debug(f'PESQ error {e}')
        score = error_value

    return score


def plot_mixpred(mixture: AudioT,
                 mixture_f: AudioF,
                 target: AudioT = None,
                 feature: Feature = None,
                 predict: Predict = None,
                 tp_title: str = '') -> plt.figure:
    from sonusai.mixture import SAMPLE_RATE

    num_plots = 2
    if feature is not None:
        num_plots += 1
    if predict is not None:
        num_plots += 1

    fig, ax = plt.subplots(num_plots, 1, constrained_layout=True, figsize=(11, 8.5))

    # Plot the waveform
    p = 0
    x_axis = np.arange(len(mixture), dtype=np.float32) / SAMPLE_RATE
    ax[p].plot(x_axis, mixture, label='Mixture', color='mistyrose')
    ax[0].set_ylabel('magnitude', color='tab:blue')
    ax[p].set_xlim(x_axis[0], x_axis[-1])
    # ax[p].set_ylim([-1.025, 1.025])
    if target is not None:  # Plot target time-domain waveform on top of mixture
        ax[0].plot(x_axis, target, label='Target', color='tab:blue')
        # ax[0].tick_params(axis='y', labelcolor=color)
    ax[p].set_title('Waveform')

    # Plot the mixture spectrogram
    p += 1
    ax[p].imshow(np.transpose(mixture_f), aspect='auto', interpolation='nearest', origin='lower')
    ax[p].set_title('Mixture')

    if feature is not None:
        p += 1
        ax[p].imshow(np.transpose(feature), aspect='auto', interpolation='nearest', origin='lower')
        ax[p].set_title('Feature')

    if predict is not None:
        p += 1
        ax[p].imshow(np.transpose(predict), aspect='auto', interpolation='nearest', origin='lower')
        ax[p].set_title('Predict ' + tp_title)

    return fig


def plot_pdb_predtruth(predict: np.ndarray,
                       truth_f: Union[np.ndarray, None] = None,
                       metric: Union[np.ndarray, None] = None,
                       tp_title: str = '') -> plt.figure:
    """ Plot predict and optionally truth and a metric in power db, e.g. applies 10*log10(predict)"""

    num_plots = 2
    if truth_f is not None:
        num_plots += 1

    fig, ax = plt.subplots(num_plots, 1, constrained_layout=True, figsize=(11, 8.5))

    # Plot the predict spectrogram
    p = 0
    tmp = 10 * np.log10(predict.transpose() + np.finfo(np.float32).eps)
    ax[p].imshow(tmp, aspect='auto', interpolation='nearest', origin='lower')
    ax[p].set_title('Predict')

    if truth_f is not None:
        p += 1
        tmp = 10 * np.log10(truth_f.transpose() + np.finfo(np.float32).eps)
        ax[p].imshow(tmp, aspect='auto', interpolation='nearest', origin='lower')
        ax[p].set_title('Truth')

    # Plot the predict avg, and optionally truth avg and metric lines
    pred_avg = 10 * np.log10(np.mean(predict, axis=-1) + np.finfo(np.float32).eps)
    p += 1
    x_axis = np.arange(len(pred_avg), dtype=np.float32)  # / SAMPLE_RATE
    ax[p].plot(x_axis, pred_avg, color='black', linestyle='dashed', label='Predict mean over freq.')
    ax[p].set_ylabel('mean db', color='black')
    ax[p].set_xlim(x_axis[0], x_axis[-1])
    if truth_f is not None:
        truth_avg = 10 * np.log10(np.mean(truth_f, axis=-1) + np.finfo(np.float32).eps)
        ax[p].plot(x_axis, truth_avg, color='green', linestyle='dashed', label='Truth mean over freq.')

    if metric is not None:  # instantiate 2nd y-axis that shares the same x-axis
        ax2 = ax[p].twinx()
        color2 = 'red'
        ax2.plot(x_axis, metric, color=color2, label='sig distortion (mse db)')
        ax2.set_xlim(x_axis[0], x_axis[-1])
        ax2.set_ylim([0, np.max(metric)])
        ax2.set_ylabel('spectral distortion (mse db)', color=color2)
        ax2.tick_params(axis='y', labelcolor=color2)
        ax[p].set_title('SNR and SNR mse (mean over freq. db)')
    else:
        ax[p].set_title('SNR (mean over freq. db)')
        # ax[0].tick_params(axis='y', labelcolor=color)
    return fig


def plot_epredtruth(predict: np.ndarray,
                    predict_wav: np.ndarray,
                    truth_f: Union[np.ndarray, None] = None,
                    truth_wav: Union[np.ndarray, None] = None,
                    metric: Union[np.ndarray, None] = None,
                    tp_title: str = '') -> plt.figure:
    """ Plot predict spectrogram and waveform and optionally truth and a metric)"""

    num_plots = 2
    if truth_f is not None:
        num_plots += 1
    if metric is not None:
        num_plots += 1

    fig, ax = plt.subplots(num_plots, 1, constrained_layout=True, figsize=(11, 8.5))

    # Plot the predict spectrogram
    p = 0
    ax[p].imshow(predict.transpose(), aspect='auto', interpolation='nearest', origin='lower')
    ax[p].set_title('Predict')

    if truth_f is not None:
        p += 1
        ax[p].imshow(truth_f.transpose(), aspect='auto', interpolation='nearest', origin='lower')
        ax[p].set_title('Truth')

    # Plot the predict wav, and optionally truth avg and metric lines
    p += 1
    x_axis = np.arange(len(predict_wav), dtype=np.float32)  # / SAMPLE_RATE
    ax[p].plot(x_axis, predict_wav, color='black', linestyle='dashed', label='Speech Estimate')
    ax[p].set_ylabel('Amplitude', color='black')
    ax[p].set_xlim(x_axis[0], x_axis[-1])
    if truth_wav is not None:
        ntrim = len(truth_wav) - len(predict_wav)
        if ntrim > 0:
            truth_wav = truth_wav[0:-ntrim]
        ax[p].plot(x_axis, truth_wav, color='green', linestyle='dashed', label='True Target')

    # Plot the metric lines
    if metric is not None:
        p += 1
        x_axis = np.arange(len(metric), dtype=np.float32)  # / SAMPLE_RATE
        ax[p].plot(x_axis, metric, color='red', label='Target LogErr')
        ax[p].set_ylabel('log error db', color='red')
        ax[p].set_xlim(x_axis[0], x_axis[-1])
        ax[p].set_ylim([-0.01, np.max(metric) + .01])

    return fig


def main():
    from docopt import docopt

    import sonusai
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

    verbose = args['--verbose']
    mixids = args['--mixid']
    predict_name = args['--ploc']
    truth_est_mode = args['--truth-est-mode']
    plot_enable = args['--plot-enable']
    disable_wav = args['--disable-wav']
    input_name = args['INPUT']

    from os.path import basename
    from os.path import join
    from os.path import splitext
    import pandas as pd

    import h5py

    from sonusai import create_file_handler
    from sonusai import initial_log_messages
    from sonusai import update_console_handler
    from sonusai.mixture import MixtureDatabase
    from sonusai.utils import int16_to_float
    from sonusai.utils import float_to_int16
    from sonusai.utils import stacked_complex_imag
    from sonusai.utils import stacked_complex_real
    from sonusai.utils import unstack_complex
    from sonusai.utils import write_wav

    # Setup logging file
    create_file_handler(join(predict_name, 'calc_metric_spenh_targetf.log'))
    update_console_handler(verbose)
    initial_log_messages('calc_metric_spenh_targetf')

    mixdb = MixtureDatabase(input_name)
    mixids = mixdb.mixids_to_list(mixids)
    logger.info(f'Found {len(mixids)} mixtures with {mixdb.num_classes} classes from {input_name}')

    all_metrics_table_1 = None
    all_metrics_table_2 = None
    all_metrics_table_3 = None
    all_metrics_table_4 = None

    target_truth_est_complex = None
    noise_truth_est_complex = None
    target_truth_est_audio = None
    noise_truth_est_audio = None
    lerr_tgtr_frame = None
    lerr_ntr_frame = None
    mtab_snr_summary_tr = None
    mtab_snr_summary_emtr = None

    for mi in mixids:
        # 1) Collect true target, noise, mixture data
        target = mixdb.mixture_target(mi)
        noise = mixdb.mixture_noise(mi)
        mixture = mixdb.mixture_mixture(mi, target=target, noise=noise)
        feature, truth_f = mixdb.mixture_ft(mi, mixture=mixture)
        mixture_f = mixdb.mixture_mixture_f(mi, mixture=mixture)
        target_f = mixdb.mixture_target_f(mi, target=target)
        noise_f = mixdb.mixture_noise_f(mi, noise=noise)

        target = int16_to_float(target)
        noise = int16_to_float(noise)
        mixture = int16_to_float(mixture)

        # 2)  Read predict data
        output_name = join(predict_name, mixdb.mixtures[mi].name)
        base_name = splitext(output_name)[0]
        with h5py.File(output_name, 'r') as f:
            predict = np.array(f['predict'])

        # 3) Extraction - for target_f truth, simply unstack as predict is estimating the target
        predict_complex = unstack_complex(predict)
        truth_f_complex = unstack_complex(truth_f)
        noise_est_complex = mixture_f - predict_complex
        target_est_wav = mixdb.inverse_transform(predict_complex)
        noise_est_wav = mixdb.inverse_transform(noise_est_complex)

        if truth_est_mode is True:
            # estimates using truth instead of prediction
            target_truth_est_complex = truth_f_complex
            noise_truth_est_complex = mixture_f - target_truth_est_complex
            target_truth_est_audio = mixdb.inverse_transform(target_truth_est_complex)
            noise_truth_est_audio = mixdb.inverse_transform(noise_truth_est_complex)

        # 4) Metrics
        # Mean absolute error (real and imag)
        mape_r_tg, mape_r_tg_bin, mape_r_tg_frame = mean_abs_percentage_error(stacked_complex_real(truth_f),
                                                                              stacked_complex_real(predict))
        mape_i_tg, mape_i_tg_bin, mape_i_tg_frame = mean_abs_percentage_error(stacked_complex_imag(truth_f),
                                                                              stacked_complex_imag(predict))
        cmape_tg = (mape_r_tg / 2) + (mape_i_tg / 2)
        cmape_tg_bin = (mape_r_tg_bin / 2) + (mape_i_tg_bin / 2)
        cmape_tg_frame = (mape_r_tg_frame / 2) + (mape_i_tg_frame / 2)

        # Target/Speech logerr - PSD estimation accuracy symmetric mean log-spectral distortion
        lerr_tg, lerr_tg_bin, lerr_tg_frame = log_error(truth_f_complex, predict_complex)
        # Noise logerr - PSD estimation accuracy
        lerr_n, lerr_n_bin, lerr_n_frame = log_error(noise_f, noise_est_complex)

        # Speech intelligibility measure - PESQ
        pesq_speech = calculate_pesq(target, target_est_wav)
        pesq_mixture = calculate_pesq(target, mixture)
        pesq_impr = pesq_speech - pesq_mixture  # pesq improvement
        pesq_impr_pc = pesq_impr / (pesq_mixture + np.finfo(np.float32).eps) * 100  # pesq improvement %

        # 5) Save per mixture metric results
        # Single row in table of scalar metrics per mixture
        mtable1_col = ['MXSNR', 'MXPESQ', 'PESQ', 'PESQi', 'PESQi%', 'SPCMAPE', 'SPLERR', 'NLERR', 'SPFILE', 'NFILE']
        ti = mixdb.mixtures[mi].target_file_index[0]
        ni = mixdb.mixtures[mi].noise_file_index
        metr1 = [mixdb.mixtures[mi].snr, pesq_mixture, pesq_speech, pesq_impr, pesq_impr_pc, cmape_tg, lerr_tg,
                 lerr_n, basename(mixdb.targets[ti].name), basename(mixdb.noises[ni].name)]
        mtab1 = pd.DataFrame([metr1], columns=mtable1_col, index=[mi])
        # Stats of per frame estimation metrics
        efs_table2_col = ['Max', 'Min', 'Avg', 'Median']
        efs_table2_row = ['SPCMAPE', 'SPLERR', 'NLERR']
        metr2 = [[np.max(cmape_tg_frame), np.min(cmape_tg_frame), np.mean(cmape_tg_frame), np.median(cmape_tg_frame)],
                 [np.max(lerr_tg_frame), np.min(lerr_tg_frame), np.mean(lerr_tg_frame), np.median(lerr_tg_frame)],
                 [np.max(lerr_n_frame), np.min(lerr_n_frame), np.mean(lerr_n_frame), np.median(lerr_n_frame)]]
        mtab2 = pd.DataFrame(metr2, columns=efs_table2_col, index=efs_table2_row)

        mtab2flat_col = ['MXSNR', 'SPCMAPE Max', 'SPCMAPE Min', 'SPCMAPE Avg', 'SPCMAPE Median',
                         'SPLERR Max', 'SPLERR Min', 'SPLERR Avg', 'SPLERR Median',
                         'NLERR Max', 'NLERR Min', 'NLERR Avg', 'NLERR Median']
        tmp = np.insert(np.array(metr2), 0, mixdb.mixtures[mi].snr).reshape(1, 13)
        mtab2_flat = pd.DataFrame(tmp, columns=mtab2flat_col, index=[mi])

        if all_metrics_table_1 is None:
            # create the table if it doesn't exist
            all_metrics_table_1 = mtab1
        else:
            # Update all mixture table
            all_metrics_table_1 = pd.concat([all_metrics_table_1, mtab1])

        if all_metrics_table_2 is None:
            # create the table if it doesn't exist
            all_metrics_table_2 = mtab2_flat
        else:
            # Update all mixture table
            all_metrics_table_2 = pd.concat([all_metrics_table_2, mtab2_flat])

        metric_name = base_name + '-spenh-metrics.txt'
        with open(metric_name, 'w') as f:
            print('Speech enhancement metrics:', file=f)
            print(mtab1.round(2).to_string(), file=f)
            print('', file=f)
            print(f'Extraction statistics over {mixture_f.shape[0]} frames:', file=f)
            print(mtab2.round(2).to_string(), file=f)
            print('', file=f)
            print(f'Target path: {mixdb.targets[ti].name}', file=f)
            print(f'Noise path: {mixdb.noises[ni].name}', file=f)
            # print(f'PESQ improvement: {pesq_impr:0.2f}, {pesq_impr_pc:0.1f}%', file=f)

        if truth_est_mode is True:
            # metrics of estimates using truth instead of prediction
            lerr_tgtr, lerr_tgtr_bin, lerr_tgtr_frame = log_error(truth_f_complex, target_truth_est_complex)
            lerr_ntr, lerr_ntr_bin, lerr_ntr_frame = log_error(noise_f, noise_truth_est_complex)
            pesq_speechtr = calculate_pesq(target, target_truth_est_audio)
            pesq_impr_sptr = pesq_speechtr - pesq_mixture  # pesq improvement
            pesq_impr_pctr = pesq_impr_sptr / (pesq_mixture + np.finfo(np.float32).eps) * 100  # pesq improvement %

            mtable3_col = ['MXSNR', 'MXPESQ', 'PESQ', 'PESQi', 'PESQi%', 'SPLERR', 'NLERR']
            metr3 = [mixdb.mixtures[mi].snr, pesq_mixture, pesq_speechtr, pesq_impr_sptr, pesq_impr_pctr, lerr_tgtr,
                     lerr_ntr]
            mtab3 = pd.DataFrame([metr3], columns=mtable3_col, index=[mi])

            # Stats of per frame estimation metrics
            efs_table4_col = ['Max', 'Min', 'Avg', 'Median']
            efs_table4_row = ['SPLERR', 'NLERR']
            metr4 = [[np.max(lerr_tgtr_frame), np.min(lerr_tgtr_frame), np.mean(lerr_tgtr_frame),
                      np.median(lerr_tgtr_frame)],
                     [np.max(lerr_ntr_frame), np.min(lerr_ntr_frame), np.mean(lerr_ntr_frame),
                      np.median(lerr_ntr_frame)]]
            mtab4 = pd.DataFrame(metr4, columns=efs_table4_col, index=efs_table4_row)

            # Append extraction metrics to metrics file:
            with open(metric_name, 'a') as f:
                print('', file=f)
                print('Speech enhancement metrics of extraction method using truth:', file=f)
                print(mtab3.round(2).to_string(), file=f)
                print('', file=f)
                print('Extraction (using Truth) statistics over frames:', file=f)
                print(mtab4.round(2).to_string(), file=f)

            # Append to all mixture table
            mtab4flat_col = ['MXSNR', 'SPLERR Max', 'SPLERR Min', 'SPLERR Avg', 'SPLERR Median',
                             'NLERR Max', 'NLERR Min', 'NLERR Avg', 'NLERR Median']
            tmp = np.insert(np.array(metr4), 0, mixdb.mixtures[mi].snr).reshape(1, 9)  # Insert MXSNR
            mtab4_flat = pd.DataFrame(tmp, columns=mtab4flat_col, index=[mi])

            if all_metrics_table_3 is None:
                all_metrics_table_3 = mtab3
            else:
                all_metrics_table_3 = pd.concat([all_metrics_table_3, mtab3])

            if all_metrics_table_4 is None:
                all_metrics_table_4 = mtab4_flat
            else:
                all_metrics_table_4 = pd.concat([all_metrics_table_4, mtab4_flat])

        # 7) write wav files
        if disable_wav is False:
            write_wav(name=base_name + '-mixture.wav', audio=float_to_int16(mixture))
            write_wav(name=base_name + '-target-est.wav', audio=float_to_int16(target_est_wav))
            write_wav(name=base_name + '-noise-est.wav', audio=float_to_int16(noise_est_wav))
            write_wav(name=base_name + '-target.wav', audio=float_to_int16(target))
            write_wav(name=base_name + '-noise.wav', audio=float_to_int16(noise))
            # debug code to test for perfect reconstruction of the extraction method
            # note both 75% olsa-hanns and 50% olsa-hann modes checked to have perfect reconstruction
            # target_r = mixdb.inverse_transform(target_f)
            # noise_r = mixdb.inverse_transform(noise_f)
            # _write_wav(name=base_name + '-target_r.wav', audio=float_to_int16(target_r))
            # _write_wav(name=base_name + '-noise_r.wav', audio=float_to_int16(noise_r)) # chk perfect rec
            if truth_est_mode is True:
                write_wav(name=base_name + '-target-trest.wav', audio=float_to_int16(target_truth_est_audio))
                write_wav(name=base_name + '-noise-trest.wav', audio=float_to_int16(noise_truth_est_audio))

        # 8) Write out plot file
        if plot_enable is True:
            from matplotlib.backends.backend_pdf import PdfPages
            plot_fname = base_name + '-spenh-plots.pdf'

            # Reshape feature to eliminate overlap redundancy for easier to understand spectrogram view
            # Original size (frames, stride, num_bands), decimates in stride dimension only if step is > 1
            # Reshape to get frames*decimated_stride, num_bands
            step = int(mixdb.feature_samples / mixdb.feature_step_samples)
            if feature.ndim != 3:
                raise SonusAIError(f'feature does not have 3 dimensions: frames, stride, num_bands')

            # for feature cn*00n**
            feat_sgram = unstack_complex(feature)
            feat_sgram = 20 * np.log10(abs(feat_sgram) + np.finfo(np.float32).eps)
            feat_sgram = feat_sgram[:, -step:, :]  # decimate,  Fx1xB
            feat_sgram = np.reshape(feat_sgram, (feat_sgram.shape[0] * feat_sgram.shape[1], feat_sgram.shape[2]))

            with PdfPages(plot_fname) as pdf:
                # page1 we always have a mixture and prediction, target optional if truth provided
                tfunc_name = mixdb.targets[0].truth_settings[0].function  # first target, assumes all have same
                if tfunc_name == 'mapped_snr_f':  # leave as unmapped snr
                    predplot = predict
                    tfunc_name = mixdb.targets[0].truth_settings[0].function
                elif tfunc_name == 'target_f':
                    predplot = 20 * np.log10(abs(predict_complex) + np.finfo(np.float32).eps)
                else:
                    predplot = 10 * np.log10(predict + np.finfo(np.float32).eps)  # use db scale
                    tfunc_name = tfunc_name + ' (db)'

                mixspec = 20 * np.log10(abs(mixture_f) + np.finfo(np.float32).eps)
                pdf.savefig(plot_mixpred(mixture=mixture,
                                         mixture_f=mixspec,
                                         target=target,
                                         feature=feat_sgram,
                                         predict=predplot,
                                         tp_title=tfunc_name))

                # ----- page 2, plot unmapped predict, opt truth reconstructed and line plots of mean-over-f
                # pdf.savefig(plot_pdb_predtruth(predict=pred_snr_f, tp_title='predict snr_f (db)'))

                # page3
                tg_spec = 20 * np.log10(abs(target_f) + np.finfo(np.float32).eps)
                tg_est_spec = 20 * np.log10(abs(predict_complex) + np.finfo(np.float32).eps)
                # n_spec = np.reshape(n_spec,(n_spec.shape[0] * n_spec.shape[1], n_spec.shape[2]))
                pdf.savefig(plot_epredtruth(predict=tg_est_spec,
                                            predict_wav=target_est_wav,
                                            truth_f=tg_spec,
                                            truth_wav=target,
                                            metric=lerr_tg_frame,
                                            tp_title='speech_est'))

                # page 4 noise extraction
                n_spec = 20 * np.log10(abs(noise_f) + np.finfo(np.float32).eps)
                n_est_spec = 20 * np.log10(abs(noise_est_complex) + np.finfo(np.float32).eps)
                pdf.savefig(plot_epredtruth(predict=n_est_spec,
                                            predict_wav=noise_est_wav,
                                            truth_f=n_spec,
                                            truth_wav=noise,
                                            metric=lerr_n_frame,
                                            tp_title='noise_est'))

                # page 5 truth-based speech extraction
                if truth_est_mode is True:
                    tg_trest_spec = 20 * np.log10(abs(target_truth_est_complex) + np.finfo(np.float32).eps)
                    pdf.savefig(plot_epredtruth(predict=tg_trest_spec,
                                                predict_wav=target_truth_est_audio,
                                                truth_f=tg_spec,
                                                truth_wav=target,
                                                metric=lerr_tgtr_frame,
                                                tp_title='speech_trest'))

                    # page 6 truth-based noise extraction
                    n_trest_spec = 20 * np.log10(abs(noise_truth_est_complex) + np.finfo(np.float32).eps)
                    pdf.savefig(plot_epredtruth(predict=n_trest_spec,
                                                predict_wav=noise_truth_est_audio,
                                                truth_f=n_spec,
                                                truth_wav=noise,
                                                metric=lerr_ntr_frame,
                                                tp_title='truth-based noise estimate'))

            plt.close('all')
            logger.info(f'Wrote plots to {plot_fname}')

    # 9) Done with mixtures, write out summary metrics
    # Calculate SNR summary for
    all_mtab1_sorted = all_metrics_table_1.sort_values(by=['MXSNR', 'SPFILE'])
    all_mtab2_sorted = all_metrics_table_2.sort_values(by=['MXSNR'])
    mtab_snr_summary = all_mtab1_sorted.query('MXSNR==' + str(mixdb.snrs[0])).mean(numeric_only=True).to_frame().T
    mtab_snr_summary_em = all_mtab2_sorted.query('MXSNR==' + str(mixdb.snrs[0])).mean(numeric_only=True).to_frame().T
    for snri in range(1, len(mixdb.snrs)):
        mtab_snr_summary = pd.concat([mtab_snr_summary,
                                      all_mtab1_sorted.query('MXSNR==' + str(mixdb.snrs[snri])).mean(
                                          numeric_only=True).to_frame().T])
        mtab_snr_summary_em = pd.concat([mtab_snr_summary_em,
                                         all_mtab2_sorted.query('MXSNR==' + str(mixdb.snrs[snri])).mean(
                                             numeric_only=True).to_frame().T])

    if truth_est_mode is True:
        all_mtab3_sorted = all_metrics_table_3.sort_values(by=['MXSNR'])
        all_mtab4_sorted = all_metrics_table_4.sort_values(by=['MXSNR'])
        mtab_snr_summary_tr = all_mtab3_sorted.query('MXSNR==' + str(mixdb.snrs[0])).mean(
            numeric_only=True).to_frame().T
        mtab_snr_summary_emtr = all_mtab4_sorted.query('MXSNR==' + str(mixdb.snrs[0])).mean(
            numeric_only=True).to_frame().T
        for snri in range(1, len(mixdb.snrs)):
            mtab_snr_summary_tr = pd.concat([mtab_snr_summary_tr,
                                             all_mtab3_sorted.query('MXSNR==' + str(mixdb.snrs[snri])).mean(
                                                 numeric_only=True).to_frame().T])
            mtab_snr_summary_emtr = pd.concat([mtab_snr_summary_emtr,
                                               all_mtab4_sorted.query('MXSNR==' + str(mixdb.snrs[snri])).mean(
                                                   numeric_only=True).to_frame().T])

    num_mix = len(mixids)
    if num_mix > 1:
        metric_summary_fname = join(predict_name, 'spenh-metrics-summary.txt')
        with open(metric_summary_fname, 'w') as f:
            print(f'Speech enhancement metrics avg over each SNR:', file=f)
            print(mtab_snr_summary.round(2).to_string(), file=f)
            print('', file=f)
            print(f'Extraction statistics stats avg over each SNR:', file=f)
            print(mtab_snr_summary_em.round(2).to_string(), file=f)
            print('', file=f)

            print(f'Speech enhancement metrics stats over all {num_mix} mixtures:', file=f)
            print(all_metrics_table_1.describe().round(2).to_string(), file=f)
            print('', file=f)
            print(f'Extraction statistics stats over all {num_mix} mixtures:', file=f)
            print(all_metrics_table_2.describe().round(2).to_string(), file=f)
            print('', file=f)

            if truth_est_mode is True:
                print(f'Truth-based speech enhancement metrics avg over each SNR:', file=f)
                print(mtab_snr_summary_tr.round(2).to_string(), file=f)
                print('', file=f)
                print(f'Truth-based extraction statistics stats avg over each SNR:', file=f)
                print(mtab_snr_summary_emtr.round(2).to_string(), file=f)
                print('', file=f)

                print(f'Truth-based speech enhancement metrics stats over all {num_mix} mixtures:', file=f)
                print(all_metrics_table_3.describe().round(2).to_string(), file=f)
                print('', file=f)
                print(f'Truth-based extraction statistic stats over all {num_mix} mixtures:', file=f)
                print(all_metrics_table_4.describe().round(2).to_string(), file=f)
                print('', file=f)

            print('Speech enhancement metrics all-mixtures list:', file=f)
            print(all_metrics_table_1.round(2).to_string(), file=f)
            print('', file=f)
            print('Extraction statistics all-mixtures list:', file=f)
            print(all_metrics_table_2.round(2).to_string(), file=f)

            # Write summary to .csv file
            csv_name = join(predict_name, 'spenh-metrics-summary.csv')
            header_args = {
                'mode': 'a',
                'encoding': 'utf-8',
                'index': False,
                'header': False,
            }
            table_args = {
                'mode': 'a',
                'encoding': 'utf-8',
            }
            pd.DataFrame([f'Speech enhancement metrics stats over {num_mix} mixtures:']).to_csv(csv_name, **header_args)
            all_metrics_table_1.describe().round(2).to_csv(csv_name, encoding='utf-8')
            pd.DataFrame(['']).to_csv(csv_name, **header_args)

            pd.DataFrame([f'Extraction statistics stats over {num_mix} mixtures:']).to_csv(csv_name, **header_args)
            all_metrics_table_2.describe().round(2).to_csv(csv_name, **table_args)
            pd.DataFrame(['']).to_csv(csv_name, **header_args)

            if truth_est_mode is True:
                pd.DataFrame(['Truth estimation metrics stats:']).to_csv(csv_name, **header_args)
                all_metrics_table_3.describe().round(2).to_csv(csv_name, **table_args)
                pd.DataFrame(['']).to_csv(csv_name, **header_args)

                pd.DataFrame(['Truth estimation extraction statistics stats:']).to_csv(csv_name, **header_args)
                all_metrics_table_4.describe().round(2).to_csv(csv_name, **table_args)
                pd.DataFrame(['']).to_csv(csv_name, **header_args)

            csv_name = join(predict_name, 'spenh-metrics-list.csv')
            pd.DataFrame(['Speech enhancement metrics list:']).to_csv(csv_name, **header_args)
            all_metrics_table_1.round(2).to_csv(csv_name, **table_args)

            csv_name = join(predict_name, 'spenh-metrics-estats-list.csv')
            pd.DataFrame(['Extraction statistics list:']).to_csv(csv_name, **header_args)
            all_metrics_table_2.round(2).to_csv(csv_name, **table_args)

            if truth_est_mode is True:
                csv_name = join(predict_name, 'spenh-metrics-tru-list.csv')
                pd.DataFrame(['Speech enhancement metrics list:']).to_csv(csv_name, **header_args)
                all_metrics_table_3.round(2).to_csv(csv_name, **table_args)

                csv_name = join(predict_name, 'spenh-metrics-estats-tru-list.csv')
                pd.DataFrame(['Extraction statistics list:']).to_csv(csv_name, **header_args)
                all_metrics_table_4.round(2).to_csv(csv_name, **table_args)


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        exit()



from dataclasses import dataclass
from typing import Any
from typing import List
from typing import Tuple
from typing import Union

from pyaaware import FeatureGenerator

from sonusai.mixture.types import AudioF
from sonusai.mixture.types import AudioT
from sonusai.mixture.types import AudiosF
from sonusai.mixture.types import AudiosT
from sonusai.mixture.types import Augmentation
from sonusai.mixture.types import Augmentations
from sonusai.mixture.types import ClassCount
from sonusai.mixture.types import DataClassSonusAIMixin
from sonusai.mixture.types import Feature
from sonusai.mixture.types import Location
from sonusai.mixture.types import MixtureIDs
from sonusai.mixture.types import NoiseFiles
from sonusai.mixture.types import Optional
from sonusai.mixture.types import Segsnr
from sonusai.mixture.types import TargetFiles
from sonusai.mixture.types import Truth
from sonusai.mixture.types import TruthSettings


@dataclass
class MRecord(DataClassSonusAIMixin):
    name: str = None
    noise_augmentation_index: int = None
    noise_file_index: int = None
    noise_offset: int = None
    noise_snr_gain: float = None
    random_snr: Optional[bool] = None
    samples: int = None
    snr: float = None
    target_augmentation_index: List[int] = None
    target_file_index: List[int] = None
    target_gain: List[int] = None
    target_snr_gain: float = None


MRecords = List[MRecord]


@dataclass
class MixtureDatabaseConfig(DataClassSonusAIMixin):
    class_balancing: Optional[bool] = False
    class_balancing_augmentation: Optional[Augmentation] = None
    class_labels: List[str] = None
    class_weights_threshold: List[float] = None
    exhaustive_noise: Optional[bool] = True
    feature: str = None
    first_cba_index: Optional[int] = None
    mixtures: MRecords = None
    noise_augmentations: Augmentations = None
    noises: NoiseFiles = None
    num_classes: int = None
    random_snrs: Optional[List[str]] = None
    seed: Optional[int] = 0
    snrs: List[float] = None
    target_augmentations: Augmentations = None
    targets: TargetFiles = None
    truth_mutex: bool = None
    truth_reduction_function: str = None
    truth_settings: TruthSettings = None


@dataclass
class TransformConfig:
    N: int
    R: int
    ttype: str


class MixtureDatabase:
    from sonusai.mixture.types import AudiosT
    from sonusai.mixture.types import ListAudiosT

    def __init__(self, config: Union[Location, MixtureDatabaseConfig]):
        if isinstance(config, MixtureDatabaseConfig):
            self.config = config
            self.location = None
        else:
            self.load(config)
            self.location = config

        self.fg = FeatureGenerator(feature_mode=self.feature,
                                   num_classes=self.num_classes,
                                   truth_mutex=self.truth_mutex)

        self.ft_config = TransformConfig(N=self.fg.ftransform_N,
                                         R=self.fg.ftransform_R,
                                         ttype=self.fg.ftransform_ttype)

        self.eft_config = TransformConfig(N=self.fg.eftransform_N,
                                          R=self.fg.eftransform_R,
                                          ttype=self.fg.eftransform_ttype)

        self.it_config = TransformConfig(N=self.fg.itransform_N,
                                         R=self.fg.itransform_R,
                                         ttype=self.fg.itransform_ttype)

        self._raw_target_audios = None
        self._augmented_noise_audios = None

    def load(self, location: Location) -> None:
        import json
        from os.path import exists
        from os.path import isdir
        from os.path import join

        from sonusai import SonusAIError

        if not isdir(location):
            raise SonusAIError(f'{location} is not a directory')

        filename = join(location, 'mixdb.json')
        if not exists(filename):
            raise SonusAIError(f'could not find mixture database in {location}')

        with open(file=filename, mode='r', encoding='utf-8') as f:
            self.config = MixtureDatabaseConfig.from_dict(json.loads(f.read()))

    def save(self):
        from os.path import join

        json_name = join(self.location, 'mixdb.json')
        with open(file=json_name, mode='w') as file:
            file.write(self.config.to_json(indent=2))

    def forward_transform(self, audio: AudioT) -> AudioF:
        from pyaaware import ForwardTransform

        from sonusai.utils import int16_to_float

        ftf = ForwardTransform(N=self.ft_config.N, R=self.ft_config.R, ttype=self.ft_config.ttype)
        return ftf.execute_all(int16_to_float(audio)).transpose()

    def inverse_transform(self, transform: AudioF, trim: bool = True) -> AudioT:
        from pyaaware import InverseTransform

        itf = InverseTransform(N=self.it_config.N, R=self.it_config.R, ttype=self.it_config.ttype)
        audio = itf.execute_all(transform.transpose())
        if trim:
            audio = audio[itf.N - itf.R:]

        return audio

    @property
    def class_balancing(self):
        return self.config.class_balancing

    @class_balancing.setter
    def class_balancing(self, value):
        self.config.class_balancing = value

    @property
    def class_balancing_augmentation(self):
        return self.config.class_balancing_augmentation

    @class_balancing_augmentation.setter
    def class_balancing_augmentation(self, value):
        self.config.class_balancing_augmentation = value

    @property
    def class_labels(self):
        return self.config.class_labels

    @class_labels.setter
    def class_labels(self, value):
        from sonusai import SonusAIError

        if value is not None and (not isinstance(value, list) or len(value) != self.num_classes):
            raise SonusAIError(f'invalid class_labels length')
        self.config.class_labels = value

    @property
    def class_weights_threshold(self):
        return self.config.class_weights_threshold

    @class_weights_threshold.setter
    def class_weights_threshold(self, value):
        from sonusai import SonusAIError

        if len(value) not in [1, self.num_classes]:
            raise SonusAIError(f'invalid class_weights_threshold length: {len(class_weights_threshold)}')
        self.config.class_weights_threshold = value

    @property
    def exhaustive_noise(self):
        return self.config.exhaustive_noise

    @exhaustive_noise.setter
    def exhaustive_noise(self, value):
        self.config.exhaustive_noise = value

    @property
    def feature(self):
        return self.config.feature

    @feature.setter
    def feature(self, value):
        self.config.feature = value

    @property
    def transform_frame_ms(self):
        from sonusai.mixture import SAMPLE_RATE

        return float(self.ft_config.R) / float(SAMPLE_RATE / 1000)

    @property
    def feature_ms(self):
        return self.transform_frame_ms * self.fg.decimation * self.fg.stride

    @property
    def feature_samples(self):
        return self.ft_config.R * self.fg.decimation * self.fg.stride

    @property
    def feature_step_ms(self):
        return self.transform_frame_ms * self.fg.decimation * self.fg.step

    @property
    def feature_step_samples(self):
        return self.ft_config.R * self.fg.decimation * self.fg.step

    @property
    def first_cba_index(self):
        return self.config.first_cba_index

    @first_cba_index.setter
    def first_cba_index(self, value):
        self.config.first_cba_index = value

    @property
    def mixtures(self):
        return self.config.mixtures

    @mixtures.setter
    def mixtures(self, value):
        self.config.mixtures = value

    @property
    def noise_augmentations(self):
        return self.config.noise_augmentations

    @noise_augmentations.setter
    def noise_augmentations(self, value):
        self.config.noise_augmentations = value

    @property
    def noises(self):
        return self.config.noises

    @noises.setter
    def noises(self, value):
        self.config.noises = value

    @property
    def num_classes(self):
        return self.config.num_classes

    @num_classes.setter
    def num_classes(self, value):
        self.config.num_classes = value

    @property
    def random_snrs(self):
        return self.config.random_snrs

    @random_snrs.setter
    def random_snrs(self, value):
        self.config.random_snrs = value

    @property
    def seed(self):
        return self.config.seed

    @seed.setter
    def seed(self, value):
        self.config.seed = value

    @property
    def snrs(self):
        return self.config.snrs

    @snrs.setter
    def snrs(self, value):
        self.config.snrs = value

    @property
    def target_augmentations(self):
        return self.config.target_augmentations

    @target_augmentations.setter
    def target_augmentations(self, value):
        self.config.target_augmentations = value

    @property
    def targets(self):
        return self.config.targets

    @targets.setter
    def targets(self, value):
        self.config.targets = value

    @property
    def truth_mutex(self):
        return self.config.truth_mutex

    @truth_mutex.setter
    def truth_mutex(self, value):
        self.config.truth_mutex = value

    @property
    def truth_reduction_function(self):
        return self.config.truth_reduction_function

    @truth_reduction_function.setter
    def truth_reduction_function(self, value):
        self.config.truth_reduction_function = value

    @property
    def truth_settings(self):
        return self.config.truth_settings

    @truth_settings.setter
    def truth_settings(self, value):
        self.config.truth_settings = value

    @property
    def raw_target_audios(self) -> AudiosT:
        return self._raw_target_audios

    def load_raw_target_audios(self, show_progress: bool = False) -> None:
        from tqdm import tqdm

        from sonusai.mixture import read_audio
        from sonusai.utils import p_tqdm_map

        names = [target.name for target in self.targets]
        progress = tqdm(total=len(names), desc='Read target audio', disable=not show_progress)
        self._raw_target_audios = p_tqdm_map(read_audio, names, progress=progress)
        progress.close()

    @property
    def augmented_noise_audios(self) -> ListAudiosT:
        return self._augmented_noise_audios

    def load_augmented_noise_audios(self, show_progress: bool = False) -> None:
        from functools import partial

        from tqdm import tqdm

        from sonusai.mixture import read_audio
        from sonusai.utils import p_tqdm_map

        names = [noise.name for noise in self.noises]

        progress = tqdm(total=len(names), desc='Read noise audio', disable=not show_progress)
        raw_noise_audios = p_tqdm_map(read_audio, names, progress=progress)
        progress.close()

        progress = tqdm(total=len(names), desc='Augment noise audio', disable=not show_progress)
        self._augmented_noise_audios = p_tqdm_map(partial(_augment_audio, augmentations=self.noise_augmentations),
                                                  raw_noise_audios,
                                                  progress=progress)
        progress.close()

    def total_samples(self, mixids: MixtureIDs = '*') -> int:
        return sum([self.mixture_samples(mixid) for mixid in self.mixids_to_list(mixids)])

    def mixture_samples(self, mixid: int) -> int:
        return self.mixtures[mixid].samples

    def mixture_transform_frames(self, mixid: int) -> int:
        return self.mixtures[mixid].samples // self.ft_config.R

    def mixture_feature_frames(self, mixid: int) -> int:
        return self.mixtures[mixid].samples // self.feature_step_samples

    def mixids_to_list(self, mixids: MixtureIDs = None) -> List[int]:
        from sonusai import SonusAIError

        all_mixids = list(range(len(self.mixtures)))

        if mixids is None:
            return all_mixids

        if isinstance(mixids, str):
            if mixids == '*':
                return all_mixids

            try:
                return eval(f'{all_mixids}[{mixids}]')
            except NameError:
                raise SonusAIError(f'Empty mixids {mixids}')

        if isinstance(mixids, range):
            result = list(mixids)
        elif isinstance(mixids, int):
            result = [mixids]
        else:
            result = mixids

        if not all(isinstance(x, int) and 0 <= x < len(self.mixtures) for x in result):
            raise SonusAIError(f'Invalid entries in mixids of {mixids}')

        if not result:
            raise SonusAIError(f'Empty mixids {mixids}')

        return result

    def mixture_metadata(self, mixid: int) -> str:
        mrecord = self.mixtures[mixid]
        metadata = ''
        for ti in range(len(mrecord.target_file_index)):
            tfi = mrecord.target_file_index[ti]
            tai = mrecord.target_augmentation_index[ti]
            metadata += f'target {ti} name: {self.targets[tfi].name}\n'
            metadata += f'target {ti} augmentation: {self.target_augmentations[tai].to_dict()}\n'
            metadata += f'target {ti} target_gain: {mrecord.target_gain[ti]}\n'
            truth_settings = self.targets[tfi].truth_settings
            for tsi in range(len(truth_settings)):
                metadata += f'target {ti} truth index {tsi}: {truth_settings[tsi].index}\n'
                metadata += f'target {ti} truth function {tsi}: {truth_settings[tsi].function}\n'
                metadata += f'target {ti} truth config {tsi}: {truth_settings[tsi].config}\n'
        metadata += f'noise name: {self.noises[mrecord.noise_file_index].name}\n'
        metadata += f'noise augmentation: {self.noise_augmentations[mrecord.noise_augmentation_index].to_dict()}\n'
        metadata += f'snr: {mrecord.snr}\n'
        metadata += f'random_snr: {mrecord.random_snr}\n'
        metadata += f'samples: {mrecord.samples}\n'
        metadata += f'target_snr_gain: {mrecord.target_snr_gain}\n'
        metadata += f'noise_snr_gain: {mrecord.noise_snr_gain}\n'

        return metadata

    def write_mixture_metadata(self, mixid: int) -> None:
        from os.path import splitext

        with open(file=splitext(self.mixture_filename(mixid))[0] + '.txt', mode='w') as f:
            f.write(self.mixture_metadata(mixid))

    def mixture_filename(self, mixid: int) -> Location:
        from os.path import join

        return join(self.location, self.mixtures[mixid].name) if self.location is not None else None

    def check_audio_files_exist(self) -> None:
        """Walk through all the noise and target audio files in a mixture database ensuring that they exist."""
        from os.path import exists

        from sonusai import SonusAIError
        from sonusai.mixture import tokenized_expandvars

        for file_index in range(len(self.noises)):
            file_name, _ = tokenized_expandvars(self.noises[file_index].name)
            if not exists(file_name):
                raise SonusAIError(f'Could not find {file_name}')

        for file_index in range(len(self.targets)):
            file_name, _ = tokenized_expandvars(self.targets[file_index].name)
            if not exists(file_name):
                raise SonusAIError(f'Could not find {file_name}')

    def read_mixture_data(self, mixid: int, items: Union[List[str], str]) -> Any:
        from os.path import exists

        import h5py
        import numpy as np

        def _get_dataset(file: h5py.File, name: str) -> Any:
            if name in file:
                return np.array(f[name])
            return None

        if not isinstance(items, list):
            items = [items]

        if exists(self.mixture_filename(mixid)):
            with h5py.File(self.mixture_filename(mixid), 'r') as f:
                result = ([_get_dataset(f, item) for item in items])
        else:
            result = ([None for item in items])

        if len(items) == 1:
            result = result[0]

        return result

    def write_mixture_data(self, mixid: int, items: Union[List[Tuple[str, Any]], Tuple[str, Any]]) -> None:
        import h5py

        if not isinstance(items, list):
            items = [items]

        with h5py.File(self.mixture_filename(mixid), 'a') as f:
            for item in items:
                if item[0] in f:
                    del f[item[0]]
                f.create_dataset(name=item[0], data=item[1])

    def mixture_targets(self, mixid: int, force: bool = False) -> AudiosT:
        """ Get the list of augmented target audio (one per target in the mixup) for the given mixid

        mixid
            Mixture ID.
        force
            Force computing data from original sources regardless of whether or not cached data exists.
        """
        from sonusai.mixture import apply_augmentation
        from sonusai.mixture import apply_gain
        from sonusai.mixture import pad_audio_to_length
        from sonusai.mixture import read_audio

        if not force:
            targets = self.read_mixture_data(mixid, 'targets')
            if targets is not None:
                return list(targets)

        mrecord = self.mixtures[mixid]
        targets = []
        for idx in range(len(mrecord.target_file_index)):
            if self.raw_target_audios is not None:
                target = self.raw_target_audios[mrecord.target_file_index[idx]]
            else:
                target = read_audio(self.targets[mrecord.target_file_index[idx]].name)

            target = apply_augmentation(audio=target,
                                        augmentation=self.target_augmentations[mrecord.target_augmentation_index[idx]],
                                        length_common_denominator=self.feature_step_samples)

            target = apply_gain(audio=target, gain=mrecord.target_snr_gain)
            target = pad_audio_to_length(audio=target, length=mrecord.samples)
            targets.append(target)

        return targets

    def mixture_targets_f(self, mixid: int, targets: AudiosT = None) -> AudiosF:
        """ Get the list of augmented target transform (one per target in the mixup) for the given mixid

        mixid
            Mixture ID.
        targets
            List of augmented target audio (one per target in the mixup) for the given mixid. (optional)
        """
        if targets is None:
            targets = self.mixture_targets(mixid=mixid)

        return [self.forward_transform(target) for target in targets]

    def mixture_target(self, mixid: int, targets: AudiosT = None, force: bool = False) -> AudioT:
        """ Get the augmented target audio for the given mixid

        mixid
            Mixture ID.
        targets
            List of augmented target audio (one per target in the mixup) for the given mixid. (optional)
        force
            Force computing data from original sources regardless of whether or not cached data exists.
        """
        if not force and targets is not None:
            return sum(targets)

        return sum(self.mixture_targets(mixid=mixid, force=force))

    def mixture_target_f(self, mixid: int, targets: AudiosT = None, target: AudioT = None) -> AudioF:
        """ Get the augmented target transform for the given mixid

        mixid
            Mixture ID.
        targets
            List of augmented target audio (one per target in the mixup) for the given mixid. (optional)
        target
            Augmented target audio for the given mixid. (optional)
        """
        if target is None:
            target = self.mixture_target(mixid=mixid, targets=targets)

        return self.forward_transform(target)

    def mixture_noise(self, mixid: int, force: bool = False) -> AudioT:
        """ Get the augmented noise audio for the given mixid

        mixid
            Mixture ID.
        force
            Force computing data from original sources regardless of whether or not cached data exists.
        """
        from sonusai.mixture import apply_augmentation
        from sonusai.mixture import apply_gain
        from sonusai.mixture import get_next_noise
        from sonusai.mixture import read_audio

        if not force:
            noise = self.read_mixture_data(mixid, 'noise')
            if noise is not None:
                return noise

        mrecord = self.mixtures[mixid]
        if self.augmented_noise_audios is not None:
            noise = self.augmented_noise_audios[mrecord.noise_file_index][mrecord.noise_augmentation_index]
        else:
            noise = apply_augmentation(audio=read_audio(self.noises[mrecord.noise_file_index].name),
                                       augmentation=self.noise_augmentations[mrecord.noise_augmentation_index])

        noise = get_next_noise(audio=noise, offset=mrecord.noise_offset, length=mrecord.samples)
        noise = apply_gain(audio=noise, gain=mrecord.noise_snr_gain)

        return noise

    def mixture_noise_f(self, mixid: int, noise: AudioT = None) -> AudioF:
        """ Get the augmented noise transform for the given mixid

        mixid
            Mixture ID.
        noise
            Augmented noise audio for the given mixid. (optional)
        """
        if noise is None:
            noise = self.mixture_noise(mixid=mixid)

        return self.forward_transform(noise)

    def mixture_mixture(self,
                        mixid: int,
                        targets: AudiosT = None,
                        target: AudioT = None,
                        noise: AudioT = None,
                        force: bool = False) -> AudioT:
        """ Get the mixture audio for the given mixid

        mixid
            Mixture ID.
        targets
            List of augmented target audio (one per target in the mixup) for the given mixid. (optional)
        target
            Augmented target audio for the given mixid. (optional)
        noise
            Augmented noise audio for the given mixid. (optional)
        force
            Force computing data from original sources regardless of whether or not cached data exists.
        """
        if not force:
            mixture = self.read_mixture_data(mixid, 'mixture')
            if mixture is not None:
                return mixture

        if target is None:
            target = self.mixture_target(mixid=mixid, targets=targets)

        if noise is None:
            noise = self.mixture_noise(mixid=mixid)

        mixture = target + noise

        return mixture

    def mixture_mixture_f(self, mixid: int,
                          targets: AudiosT = None,
                          target: AudioT = None,
                          noise: AudioT = None,
                          mixture: AudioT = None) -> AudioF:
        """ Get the mixture transform for the given mixid

        mixid
            Mixture ID.
        targets
            List of augmented target audio (one per target in the mixup) for the given mixid. (optional)
        target
            Augmented target audio for the given mixid. (optional)
        noise
            Augmented noise audio for the given mixid. (optional)
        mixture
            Mixture audio for the given mixid. (optional)
        """
        if mixture is None:
            mixture = self.mixture_mixture(mixid=mixid, targets=targets, target=target, noise=noise)

        return self.forward_transform(mixture)

    def mixture_truth_t(self,
                        mixid: int,
                        targets: AudiosT = None,
                        noise: AudioT = None,
                        force: bool = False) -> Truth:
        """ Get the truth_t data for the given mixid

        mixid
            Mixture ID.
        targets
            List of augmented target audio (one per target in the mixup) for the given mixid. (optional)
        noise
            Augmented noise audio for the given mixid. (optional)
        force
            Force computing data from original sources regardless of whether or not cached data exists.
        """
        import numpy as np

        from sonusai import SonusAIError
        from sonusai.mixture import TruthFunctionConfig
        from sonusai.mixture import truth_function

        if not force:
            truth_t = self.read_mixture_data(mixid, 'truth_t')
            if truth_t is not None:
                return truth_t

        if targets is None:
            targets = self.mixture_targets(mixid=mixid)

        if noise is None:
            noise = self.mixture_noise(mixid=mixid)

        mrecord = self.mixtures[mixid]
        if len(targets) != len(mrecord.target_file_index):
            raise SonusAIError('Number of target audio entries does not match number of targets')

        if not all(len(target_audio) == len(noise) for target_audio in targets):
            raise SonusAIError('Lengths of target audio do not match length of noise audio')

        truth_t = np.zeros((self.mixture_samples(mixid), self.num_classes), dtype=np.float32)
        for idx in range(len(targets)):
            for truth_setting in self.targets[mrecord.target_file_index[idx]].truth_settings:
                config = TruthFunctionConfig(
                    feature=self.feature,
                    index=truth_setting.index,
                    function=truth_setting.function,
                    config=truth_setting.config,
                    num_classes=self.num_classes,
                    mutex=self.truth_mutex,
                    target_gain=mrecord.target_gain[idx] * mrecord.target_snr_gain
                )
                truth_t += truth_function(target_audio=targets[idx], noise_audio=noise, config=config)

        return truth_t

    def mixture_segsnr_t(self,
                         mixid: int,
                         targets: AudiosT = None,
                         target: AudioT = None,
                         noise: AudioT = None,
                         force: bool = False) -> Segsnr:
        """ Get the segsnr_t data for the given mixid

        mixid
            Mixture ID.
        targets
            List of augmented target audio (one per target in the mixup) for the given mixid. (optional)
        target
            Augmented target audio for the given mixid. (optional)
        noise
            Augmented noise audio for the given mixid. (optional)
        force
            Force computing data from original sources regardless of whether or not cached data exists.
        """
        import numpy as np

        from pyaaware import ForwardTransform

        from sonusai.utils import int16_to_float

        if not force:
            segsnr_t = self.read_mixture_data(mixid, 'segsnr_t')
            if segsnr_t is not None:
                return segsnr_t

        if target is None:
            target = self.mixture_target(mixid=mixid, targets=targets)

        if noise is None:
            noise = self.mixture_noise(mixid=mixid)

        fft = ForwardTransform(N=self.ft_config.N, R=self.ft_config.R, ttype=self.ft_config.ttype)

        segsnr_t = np.empty(self.mixture_samples(mixid), dtype=np.float32)

        frame = 0
        for offset in range(0, self.mixture_samples(mixid), self.ft_config.R):
            indices = slice(offset, offset + self.ft_config.R)

            target_energy = fft.energy_t(int16_to_float(target[indices]))
            noise_energy = fft.energy_t(int16_to_float(noise[indices]))

            if noise_energy == 0:
                snr = np.float32(np.inf)
            else:
                snr = np.float32(target_energy / noise_energy)

            segsnr_t[indices] = snr

        return segsnr_t

    def mixture_segsnr(self,
                       mixid: int,
                       segsnr_t: Segsnr = None,
                       targets: AudiosT = None,
                       target: AudioT = None,
                       noise: AudioT = None,
                       force: bool = False) -> Segsnr:
        """ Get the segsnr data for the given mixid

        mixid
            Mixture ID.
        segsnr_t
            segsnr_t data for the given mixid. (optional)
        targets
            List of augmented target audio (one per target in the mixup) for the given mixid. (optional)
        target
            Augmented target audio for the given mixid. (optional)
        noise
            Augmented noise audio for the given mixid. (optional)
        force
            Force computing data from original sources regardless of whether or not cached data exists.
        """
        if not force:
            segsnr = self.read_mixture_data(mixid, 'segsnr')
            if segsnr is not None:
                return segsnr

            segsnr_t = self.read_mixture_data(mixid, 'segsnr_t')
            if segsnr_t is not None:
                return segsnr_t[0::self.ft_config.R]

        if segsnr_t is None:
            segsnr_t = self.mixture_segsnr_t(mixid=mixid, targets=targets, target=target, noise=noise)

        segsnr = segsnr_t[0::self.ft_config.R]

        return segsnr

    def mixture_ft(self,
                   mixid: int,
                   targets: AudiosT = None,
                   target: AudioT = None,
                   noise: AudioT = None,
                   mixture_f: AudioF = None,
                   mixture: AudioT = None,
                   truth_t: Truth = None,
                   force: bool = False) -> (Feature, Truth):
        """ Get the feature, truth_f data for the given mixid

        mixid
            Mixture ID.
        targets
            List of augmented target audio (one per target in the mixup) for the given mixid. (optional)
        target
            Augmented target audio for the given mixid. (optional)
        noise
            Augmented noise audio for the given mixid. (optional)
        mixture
            Mixture audio for the given mixid. (optional)
        truth_t
            truth_t for the given mixid. (optional)
        force
            Force computing data from original sources regardless of whether or not cached data exists.
        """
        import numpy as np

        from sonusai.mixture import truth_reduction

        if not force:
            feature, truth_f = self.read_mixture_data(mixid, ['feature', 'truth_f'])
            if feature is not None and truth_f is not None:
                return feature, truth_f

        if mixture_f is None:
            mixture_f = self.mixture_mixture_f(mixid=mixid,
                                               targets=targets,
                                               target=target,
                                               noise=noise,
                                               mixture=mixture)

            if truth_t is None:
                truth_t = self.mixture_truth_t(mixid=mixid, targets=targets, noise=noise)

            mrecord = self.mixtures[mixid]

            transform_frames = self.mixture_transform_frames(mixid)
            feature_frames = self.mixture_feature_frames(mixid)

            if truth_t is None:
                truth_t = np.zeros((self.mixture_samples(mixid), self.num_classes), dtype=np.float32)

            feature = np.empty((feature_frames, self.fg.stride, self.fg.num_bands), dtype=np.float32)
            truth_f = np.empty((feature_frames, self.num_classes), dtype=np.complex64)

            feature_frame = 0
            for transform_frame in range(transform_frames):
                indices = slice(transform_frame * self.ft_config.R,
                                (transform_frame + 1) * self.ft_config.R)
                self.fg.execute(mixture_f[transform_frame],
                                truth_reduction(truth_t[indices], self.truth_reduction_function))

                if self.fg.eof():
                    feature[feature_frame] = self.fg.feature()
                    truth_f[feature_frame] = self.fg.truth()
                    feature_frame += 1

            if np.isreal(truth_f).all():
                truth_f = np.float32(np.real(truth_f))

            return feature, truth_f

        def mixture_feature(self,
                            mixid: int,
                            targets: AudiosT = None,
                            noise: AudioT = None,
                            mixture: AudioT = None,
                            truth_t: Truth = None,
                            force: bool = False) -> Feature:
            """ Get the feature data for the given mixid

            mixid
                Mixture ID.
            targets
                List of augmented target audio (one per target in the mixup) for the given mixid. (optional)
            noise
                Augmented noise audio for the given mixid. (optional)
            mixture
                Mixture audio for the given mixid. (optional)
            truth_t
                truth_t for the given mixid. (optional)
            force
                Force computing data from original sources regardless of whether or not cached data exists.
            """
            feature, _ = self.mixture_ft(mixid=mixid,
                                         targets=targets,
                                         noise=noise,
                                         mixture=mixture,
                                         truth_t=truth_t,
                                         force=force)
            return feature

        def mixture_truth_f(self,
                            mixid: int,
                            targets: AudiosT = None,
                            noise: AudioT = None,
                            mixture: AudioT = None,
                            truth_t: Truth = None,
                            force: bool = False) -> Feature:
            """ Get the truth_f data for the given mixid

            mixid
                Mixture ID.
            targets
                List of augmented target audio (one per target in the mixup) for the given mixid. (optional)
            noise
                Augmented noise audio for the given mixid. (optional)
            mixture
                Mixture audio for the given mixid. (optional)
            truth_t
                truth_t for the given mixid. (optional)
            force
                Force computing data from original sources regardless of whether or not cached data exists.
            """
            _, truth_f = self.mixture_ft(mixid=mixid,
                                         targets=targets,
                                         noise=noise,
                                         mixture=mixture,
                                         truth_t=truth_t,
                                         force=force)
            return truth_f

    def mixture_class_count(self,
                            mixid: int,
                            targets: AudiosT = None,
                            noise: AudioT = None,
                            truth_t: Truth = None) -> ClassCount:
        """ Compute the number of samples for which each truth index is active for the given mixid

        targets
            List of augmented target audio (one per target in the mixup) for the given mixid. (optional)
        noise
            Augmented noise audio for the given mixid. (optional)
        mixture
            Mixture audio for the given mixid. (optional)
        truth_t
            truth_t for the given mixid. (optional)
        """
        if truth_t is None:
            truth_t = self.mixture_truth_t(mixid=mixid, targets=targets, noise=noise)

        class_count = [0] * self.num_classes
        num_classes = self.num_classes
        if self.truth_mutex:
            num_classes -= 1
        for cl in range(num_classes):
            class_count[cl] = int(np.sum(truth_t[:, cl] >= self.class_weights_threshold[cl]))

        return class_count


def get_mrecords_from_mixids(mixdb: MixtureDatabase, mixids: MixtureIDs = None) -> MRecords:
    from copy import deepcopy

    return [deepcopy(mixdb.mixtures[i]) for i in mixdb.mixids_to_list(mixids)]


def _augment_audio(audio: AudioT, augmentations: Augmentations) -> AudiosT:
    from sonusai.mixture import apply_augmentation

    return [apply_augmentation(audio, augmentation) for augmentation in augmentations]



import warnings

import numpy as np

with warnings.catch_warnings():
    warnings.simplefilter('ignore')
    from keras.utils import Sequence


class KerasFromH5(Sequence):
    """Generates data for Keras from a directory containing genft data"""
    from dataclasses import dataclass

    from sonusai.mixture import MixtureDatabase
    from sonusai.mixture import MixtureIDs

    @dataclass(frozen=True)
    class BatchParams:
        from sonusai.mixture import MixtureIDs

        mixids: MixtureIDs
        offset: int
        extra: int

    def __init__(self,
                 mixdb: MixtureDatabase,
                 mixids: MixtureIDs,
                 batch_size: int,
                 timesteps: int,
                 flatten: bool,
                 add1ch: bool,
                 shuffle: bool = False):
        """Initialization"""
        from os.path import exists

        import h5py

        from sonusai import SonusAIError

        self.mixdb = mixdb
        self.mixids = self.mixdb.mixids_to_list(mixids)
        self.batch_size = batch_size
        self.timesteps = timesteps
        self.flatten = flatten
        self.add1ch = add1ch
        self.shuffle = shuffle

        for mixid in self.mixids:
            filename = self.mixdb.mixture_filename(mixid)
            if not exists(filename):
                raise SonusAIError(f'Could not find {filename}')
            with h5py.File(filename, 'r') as f:
                datasets = ['feature', 'truth_f']
                for dataset in datasets:
                    if dataset not in f:
                        raise SonusAIError(f'Could not find "{dataset}" in {filename}')

        self.stride = self.mixdb.fg.stride
        self.num_bands = self.mixdb.fg.num_bands
        self.num_classes = self.mixdb.num_classes
        self.mixture_frame_segments = None
        self.batch_frame_segments = None
        self.total_batches = None

        self._initialize_mixtures()

    def __len__(self) -> int:
        """Denotes the number of batches per epoch"""
        return self.total_batches

    def __getitem__(self, batch_index: int) -> (np.ndarray, np.ndarray):
        """Get one batch of data"""
        from sonusai.utils import reshape_inputs

        batch_params = self.batch_params[batch_index]

        result = []
        for mixid in batch_params.mixids:
            result.append(self._get_ft(mixid))

        feature = np.vstack([result[i][0] for i in range(len(result))])
        truth = np.vstack([result[i][1] for i in range(len(result))])

        if batch_params.extra > 0:
            feature = feature[batch_params.offset:-batch_params.extra]
            truth = truth[batch_params.offset:-batch_params.extra]
        else:
            feature = feature[batch_params.offset:]
            truth = truth[batch_params.offset:]

        feature, truth = reshape_inputs(feature=feature,
                                        truth=truth,
                                        batch_size=self.batch_size,
                                        timesteps=self.timesteps,
                                        flatten=self.flatten,
                                        add1ch=self.add1ch)

        return feature, truth

    def on_epoch_end(self) -> None:
        """Modification of dataset between epochs"""
        import random

        if self.shuffle:
            random.shuffle(self.mixids)
            self._initialize_mixtures()

    def _get_ft(self, mixid: int) -> (np.ndarray, np.ndarray):
        import h5py

        with h5py.File(self.mixdb.mixture_filename(mixid), 'r') as f:
            feature = np.array(f['feature'])
            truth_f = np.array(f['truth_f'])

        return feature, truth_f

    def _initialize_mixtures(self) -> None:
        from sonusai.data_generator import get_frames_per_batch
        from sonusai.data_generator.utils import get_total_batches

        frames_per_batch = get_frames_per_batch(self.batch_size, self.timesteps)
        self.total_batches = get_total_batches(mixdb=self.mixdb,
                                               mixids=self.mixids,
                                               frames_per_batch=frames_per_batch)

        # Compute mixid, offset, and extra for dataset
        # offsets and extras are needed because mixtures are not guaranteed to fall on batch boundaries.
        # When fetching a new index that starts in the middle of a sequence of mixtures, the
        # previous feature frame offset must be maintained in order to preserve the correct
        # data sequence. And the extra must be maintained in order to preserve the correct data length.
        cumulative_frames = 0
        start_mixture_index = 0
        offset = 0
        self.batch_params = []
        self.file_indices = []
        total_frames = 0
        for idx, mixid in enumerate(self.mixids):
            current_frames = self.mixdb.mixture_samples(mixid) // self.mixdb.feature_step_samples
            self.file_indices.append(slice(total_frames, total_frames + current_frames))
            total_frames += current_frames
            cumulative_frames += current_frames
            while cumulative_frames >= frames_per_batch:
                extra = cumulative_frames - frames_per_batch
                mixids = self.mixids[start_mixture_index:idx + 1]
                self.batch_params.append(self.BatchParams(mixids=mixids, offset=offset, extra=extra))
                if extra == 0:
                    start_mixture_index = idx + 1
                    offset = 0
                else:
                    start_mixture_index = idx
                    offset = current_frames - extra
                cumulative_frames = extra



from sonusai.mixture.types import AudioT


def write_wav(name: str, audio: AudioT) -> None:
    import wave

    from sonusai.mixture import CHANNEL_COUNT
    from sonusai.mixture import SAMPLE_BYTES
    from sonusai.mixture import SAMPLE_RATE

    with wave.open(name, mode='w') as f:
        f.setnchannels(CHANNEL_COUNT)
        f.setsampwidth(SAMPLE_BYTES)
        f.setframerate(SAMPLE_RATE)
        f.writeframesraw(audio)



from sonusai.mixture.mixdb import MixtureDatabase
from sonusai.mixture.types import MixtureIDs


def get_total_batches(mixdb: MixtureDatabase, mixids: MixtureIDs, frames_per_batch: int) -> (int, int):
    from sonusai import SonusAIError

    frames = mixdb.total_samples(mixids) // mixdb.feature_step_samples
    total_batches = frames // frames_per_batch

    if total_batches == 0:
        raise SonusAIError(
            f'Error: dataset only contains {frames} frames which is not enough to fill a batch size of '
            f'{frames_per_batch}. Either provide more data or decrease the batch size')

    return total_batches


def get_frames_per_batch(batch_size: int, timesteps: int) -> int:
    return batch_size if timesteps == 0 else batch_size * timesteps
